import { Archetype } from "./Archetype";
import { commandQueue } from "./util/commandQueue";
import { idGenerator } from "./util/idGenerator";
import normalizeArchetype from "./util/normalizeArchetype";
export class World {
    constructor() {
        /** An array holding all entities known to this world. */
        this.entities = new Array();
        /** An ID generator we use for assigning IDs to newly added entities. */
        this.nextId = idGenerator(1);
        /** A list of known archetypes. */
        this.archetypes = new Map();
        /* MUTATION FUNCTIONS */
        this.createEntity = (entity = {}) => {
            /* If there already is an ID, raise an error. */
            if ("id" in entity)
                throw "Attempted to add an entity that aleady had an 'id' component.";
            /* Assign an ID */
            entity.id = this.nextId();
            /* Store the entity... */
            this.entities.push(entity);
            /* ...and add it to relevant indices. */
            this.indexEntity(entity);
            return entity;
        };
        this.destroyEntity = (entity) => {
            const pos = this.entities.indexOf(entity, 0);
            /* Sanity check */
            if (pos < 0)
                return;
            /* Remove it from our global list of entities */
            this.entities.splice(pos, 1);
            /* Remove entity from all archetypes */
            for (const archetype of this.archetypes.values()) {
                archetype.removeEntity(entity);
            }
            /* Remove its id component */
            delete entity.id;
        };
        this.addComponent = (entity, name, data) => {
            if (name in entity) {
                throw `Tried to add component "${name} to an entity that already has it.`;
            }
            /* TODO: checking entity ownership like this is likely to slow us down quite a lot, so eventually we'll want something smarter here. */
            if (!this.entities.includes(entity)) {
                throw `Tried to add component "${name}" to an entity that is not managed by this world.`;
            }
            /* Add the component */
            entity[name] = data;
            /* Trigger a reindexing of the entity */
            this.indexEntity(entity);
        };
        this.removeComponent = (entity, ...components) => {
            /* TODO: checking entity ownership like this is likely to slow us down quite a lot, so eventually we'll want something smarter here. */
            if (!this.entities.includes(entity)) {
                throw `Tried to remove ${components} from an entity that is not managed by this world.`;
            }
            for (const name of components) {
                if (!(name in entity)) {
                    throw `Tried to remove component "${name} from an entity that doesn't have it.`;
                }
                delete entity[name];
            }
            this.indexEntity(entity);
        };
        /* QUEUED MUTATION FUNCTIONS */
        /**
         * A queue of commands to run.
         */
        this.queuedCommands = commandQueue();
        this.queue = {
            createEntity: (entity) => {
                this.queuedCommands.add(() => this.createEntity(entity));
                return entity;
            },
            destroyEntity: (entity) => {
                this.queuedCommands.add(() => this.destroyEntity(entity));
            },
            addComponent: (entity, name, data) => {
                this.queuedCommands.add(() => this.addComponent(entity, name, data));
            },
            removeComponent: (entity, ...names) => {
                this.queuedCommands.add(() => this.removeComponent(entity, ...names));
            },
            flush: () => {
                this.queuedCommands.flush();
            }
        };
    }
    createArchetype(...query) {
        const normalizedQuery = normalizeArchetype(typeof query[0] === "string"
            ? { all: query }
            : query[0]);
        /* We may already have an archetype representing the same query */
        const stringifiedQuery = JSON.stringify(normalizedQuery);
        if (this.archetypes.has(stringifiedQuery))
            return this.archetypes.get(stringifiedQuery);
        /* Once we reach this point, we need to create a new archetype. */
        const archetype = new Archetype(normalizedQuery);
        this.archetypes.set(stringifiedQuery, archetype);
        for (const entity of this.entities)
            archetype.indexEntity(entity);
        return archetype;
    }
    indexEntity(entity) {
        if (this.entities.includes(entity))
            for (const archetype of this.archetypes.values())
                archetype.indexEntity(entity);
    }
    /**
     * Clear the entire world by discarding all entities and indices.
     */
    clear() {
        /* Remove all entities */
        this.entities.length = 0;
        /* Remove all archetype indices */
        this.archetypes.clear();
        /* Clear queue */
        this.queuedCommands.clear();
    }
}
