import { jsx as _jsx } from "react/jsx-runtime";
import { createContext, useContext, useEffect } from "react";
import { useConstant } from "./util/useConstant";
import { useRerender } from "./util/useRerender";
/**
 * Create various React-specific hooks and components for your
 * Miniplex ECS instance.
 *
 * @param world An instance of a Miniplex ECS to use.
 */
export function createReactIntegration(world) {
    const EntityContext = createContext(null);
    /**
     * A React component to either create a new entity, or represent an existing entity so
     * it can be enhanced with additional components (see the <Component> component.)
     */
    const Entity = ({ entity: existingEntity, children }) => {
        /* Reuse the specified entity or create a new one */
        const entity = useConstant(() => existingEntity !== null && existingEntity !== void 0 ? existingEntity : world.createEntity());
        /* If the entity was freshly created, manage its presence in the ECS world. */
        useEffect(() => {
            if (existingEntity)
                return;
            return () => world.destroyEntity(entity);
        }, [entity]);
        /* Provide a context with the entity so <Component> components can be wired up. */
        return _jsx(EntityContext.Provider, Object.assign({ value: entity }, { children: children }), void 0);
    };
    /**
     * Return the current entity context.
     */
    function useEntity() {
        return useContext(EntityContext);
    }
    /**
     * Declarative declare a component on an entity.
     */
    function Component({ name, data }) {
        const entity = useEntity();
        useEffect(() => {
            world.addComponent(entity, name, data);
            return () => {
                /* The entity might already have been destroyed, so let's check. */
                if ("id" in entity) {
                    world.removeComponent(entity, name);
                }
            };
        }, [entity, name, data]);
        return null;
    }
    /**
     * Return the entities of the specified archetype and subscribe this component
     * to it, making it re-render when entities are added to or removed from it.
     */
    function useArchetype(...query) {
        const rerender = useRerender();
        const archetype = useConstant(() => world.createArchetype(...query));
        useEffect(() => {
            archetype.onEntityAdded.on(rerender);
            archetype.onEntityRemoved.on(rerender);
            /* We need to rerender at least once, because other effects might have set up
               new entities before we had a chance to register our listeners. */
            rerender();
            return () => {
                archetype.onEntityAdded.off(rerender);
                archetype.onEntityRemoved.off(rerender);
            };
        }, [archetype]);
        return archetype.entities;
    }
    return { useArchetype, useEntity, Entity, Component };
}
